#
#    "Minimal" micropython driver for Texas Instruments ADS1220
#    4-Channel 24-Bit Analog-to-Digital Converter
#
#    Can work without the DRDY_ ADS output.
#    By default uses the internal oscillator (~4.096 MHz) and the internal 2.048 V reference.
#    Conversion operates in continuous mode.
#
#    Configuration of sensor detect features and IDACs are not supported (i.e. switched off permanently).
#
#    Created 2021 by Raul Kompass.
#
#    License: The MIT License (MIT)


from utime import sleep_us, ticks_us, ticks_diff, sleep_ms

# Command definitions, used in _write_cmd()
_START_SYNC = const(0x08)
_POWERDOWN  = const(0x02)

_Cont_Conv = const(1)   # Value used in _set_sps_convmode(); we set continuous conversion permanently
_N_Cal     = const(128)   # how many reads we do in calibration of adc: 128 takes about 6 seconds to calibrate

# Explanation of combinations ain_pos, ain_neg for _set_mux() and _MUXTAB-Entries:
#
# ain_pos, ain_neg, ain_pos*4+ain_neg, _MUXTAB[ain_pos*4+ain_neg] used in _set_mux()
#    0        0             0             0xff          => not allowed
#    0        1             1             0x00          => ain_pos = A_in_0, ain_neg = A_in_1
#    0        2             2             0x01          => ain_pos = A_in_0, ain_neg = A_in_2
#    0        3             3             0x02          => ain_pos = A_in_0, ain_neg = A_in_3
#    1        0             4             0x06          => ain_pos = A_in_1, ain_neg = A_in_0
#    1        1             5             0x0c          => (V_ref_p - V_ref_n) / 4 monitor
#    1        2             6             0x03          => ain_pos = A_in_1, ain_neg = A_in_2
#    1        3             7             0x04          => ain_pos = A_in_1, ain_neg = A_in_3
#    2        0             8             0xff          => not allowed
#    2        1             9             0xff          => not allowed
#    2        2            10             0x0d          => (AV_dd - AV_ss) / 4 monitor
#    2        3            11             0x05          => ain_pos = A_in_2, ain_neg = A_in_3
#    3        0            12             0xff          => not allowed
#    3        1            13             0xff          => not allowed
#    3        2            14             0x07          => ain_pos = A_in_3, ain_neg = A_in_2
#    3        3            15             0x0e          => ain_pos and ain_neg shortened to (AV_dd + AV_ss) / 2
# other possible , not reflected in _MUXTAB are
#    0        4                                         => ain_pos = A_in_0, ain_neg = AV_ss
#    1        4                                         => ain_pos = A_in_1, ain_neg = AV_ss
#    2        4                                         => ain_pos = A_in_2, ain_neg = AV_ss
#    3        4                                         => ain_pos = A_in_3, ain_neg = AV_ss
#
# The last 4 allow for all four inputs ain_pos= 0 .. 3 being separately measured against AV_ss.
# Use this default setting in combination with pga_byp=True and gain=1, 2, or 4.
# Input voltages can then range from AVss-0.1V to AVdd+0.1V , i.e. there are
# no restrictions regarding common mode input voltage limits as with pga_byp=False.

# Table containing multiplexer codes (bits 7-4 of reg 0) for the above input combinations
_MUXTAB = b'\xff\x00\x01\x02\x06\x0c\x03\x04\xff\xff\x0d\x05\xff\xff\x07\x0e'
_SRTAB =  b'\x04\x09\x12\x23\x42\x78\xc8'  # Table containing sampling rates
_SPSMULT = [1.25, 5, 10]
# possible sampling rates depending on ADC operating mode (duty, normal or turbo) 
# operating mode:   0            1        2
#  sps-bits       (duty)    (normal)   (turbo)
#   0               5            20       40
#   1             11.25          45       90
#   2              22.5          90      180
#   3             43.75         175      350
#   4              82.5         330      660
#   5              150          600     1200
#   6              250         1000     2000

# Constants for _set_vref_fir()
# _VREF_INT          = const(0x00)
# _VREF_EXT_REF_P0N0 = const(0x40)
# _VREF_EXT_REF_P1N1 = const(0x80)
# _VREF_AVDD_AVSS    = const(0xc0)
# _FIR_NONE          = const(0x00)  # needed
# _FIR_5060Hz        = const(0x10)
# _FIR_50Hz          = const(0x20)
# _FIR_60Hz          = const(0x30)

class Ads1220:
    """
    Micropython driver for Texas Instruments ADS1220
    4-Channel 24-Bit Analog-to-Digital Converter
    """
    # default mux: a_0 - Ground
    # default sps: 20 (sample rate = 20 SPS)
    # default gain: 1, may be set from 1 to 64)
    # default pga_byp: True, allows for AVss-0.1 <= Vin <= AVdd+0.1, limits gain to 1,2,4
    # default reference: internal, vref=2.048V
    def __init__(self, spi, cs, drdy=None, ain_pos=0, ain_neg=4, sps=20, conv_mode=1, gain=1, pga_byp=True, vref=2.048):
        self.spi = spi
        self.cs = cs
        self.drdy = drdy    # the driver is programmed so, that spi_miso can be used as drdy
        self.ain_pos = ain_pos
        self.ain_neg = ain_neg
        self.vref = vref
        self.offs = 0       # this value is set in _calibrate()

        self.cs.high()      # Make sure the serial interface is reset at cs.low()
        sleep_ms(10)        # Make sure the ADC is operational after power-on (> 9 ms)
        self.cs.low()
        self._wr_cmd(0x06)  # RESET: reset registers
        sleep_us(60)        # this takes 58 us
        self._set_gain_pga(gain, pga_byp)        # sets lower half of reg 0
        self._set_mux(ain_pos, ain_neg)          # sets upper half of reg 0, use after _set_gain_pga 
        self._set_sps_convmode(sps, conv_mode, _Cont_Conv)          # sets reg 1
        self._set_vref_fir(0x00, 0x20)           # sets reg 2: _VREF_INT and _FIR_50Hz
        self._wr_reg(3, 0x02)                    # sets reg 3, sets _DRDY_Mode
        self._wr_cmd(0x08)  # START_SYNC, rec. immediately after setting cont. conv. bit in reg 2
        
        self._calibrate()   # start real self calibration, takes many ms, dep. on sample rate
        self.cs.high()

    # Write a command, cs must be low
    def _wr_cmd(self, cmd):
        self.spi.write(bytearray([cmd]))

    # Write a register, cs must be low, reg may be 0..3
    def _wr_reg(self, reg, value): 
        self.spi.write(bytearray([0x40|(reg<<2), value]))  # 0x40: WREG

    # read a register, cs must be low, reg may be 0..3
    def _rd_reg(self, reg):
        self.spi.write(bytearray([0x20|(reg<<2)]))         # 0x20: RREG
        return int.from_bytes(self.spi.read(1), 'big')
    
    # wait for data ready signal on drdy line (or on miso if _DRDY_Mode)
    def _wait_drdy(self):
        while self.drdy.value():   # we use the drdy pin, or possibly the spi miso pin
            sleep_us(50)

    def _rd_data(self, wait = True):
        if wait:
            self._wait_drdy()
        else:
            self._wr_cmd(0x10)                         # 0x10: RDATA, needed only if we do not wait for DRDY low
            #    Note: we would not even have to write this RDATA command, if we knew that DRDY is low
        data = int.from_bytes(self.spi.read(3), 'big') # we rely on (, write=0) default in spi.read()
        self.spi.write(bytearray(b'\x00'))             # write 8 bits low to make MISO go high, assuming _DRDY_Mode
        return data if data < 0x800000 else data - 0x1000000
    
    # set input gain and programmable amplifyer bypass (for gains 1,2 or 4)
    #
    # pga_byp=True with gain=1, 2, or 4 allow for input voltages
    # ranging from AVss-0.1V to AVdd+0.1V , i.e. there are no restrictions
    # regarding common mode input voltage limits as in pga_byp=False
    #    
    def _set_gain_pga(self, gain=1, pga_byp=True):
        gain = round(gain)
        if gain > 4 and pga_byp:
            raise ValueError('Gain > 4 not allowed when PGA bypassed')
        gac = -1   # compute binary code for gain, i.e. emulate gain.bit_length()-1 which we don't have
        while gain > 0 and gac < 7 or gac < 0: # gac maximum 7; gain maximum is 1<<7 = 128
            gac +=1
            gain >>= 1
        self.gain = 1 << gac  # recreate gain value from computed pga
        self._wr_reg(0, self._rd_reg(0) & 0xF0 | gac<<1 | pga_byp & 0x01)

    # set input channels, values are 0..3 for ain_0..ain_3; a_neg = 4 is for AIN_Vss = analog ground
    # not all combinations are allowed
    # special combinations:    (1,1): (V_ref_p - V_ref_n) / 4  monitor
    # of (a_pos,a_neg)         (2,2):   (AV_dd - AV_ss) / 4  monitor
    #                          (3,3): ain_pos and ain_neg shortened to (AV_dd + AV_ss) / 2 for calibration
    # use this function after _set_gain_pga(), because a possible value error depends on pga_byp setting
    #
    def _set_mux(self, ain_pos, ain_neg=4):
        if ain_neg < 4:
            mux = _MUXTAB[ain_pos*4+ain_neg]
        if ain_neg == 4:
            mux = 8 + ain_pos
        if mux > 14:
            raise ValueError('Input channels AINp={:d} AINn={:d} not allowed'.format(ain_pos, ain_neg))
        if self._rd_reg(0) & 0x01 == 0 and ain_neg == 4:
            raise ValueError('PGA not bypassed: Incompatible with input setting ain_neg==4')
        self._wr_reg(0, self._rd_reg(0) & 0x0F | (mux << 4))

    # set sampling rate parameter and operating mode, see table in comments above
    # mode: 0 -> duty (slowest), 1 -> normal, 2 -> turbo (fastest)
    # reg 2 here is set so that the internal reference of 2.048 V is used,
    # and the low-side switch and IDAC currents are off
    #
    def _set_sps_convmode(self, sps=10, conv_mode=1, cont_conv=True):
        m_b = conv_mode ^ (~conv_mode & 0x02)>>1  # maps 0, 1, 2 to 1, 0, 2        
        for s_b in range(7):  # adjust sps according to table of possible values
            if sps < _SRTAB[s_b] * _SPSMULT[conv_mode] * 1.4:
                break
        self.sps = _SRTAB[s_b] * _SPSMULT[conv_mode]
        self._wr_reg(1, (s_b<<5) | ((m_b & 0x03)<<3) | (cont_conv<<2))

    # set voltage reference and FIR filter
    # Possible arguments: _VREF_INT, _VREF_EXT_REF_P0N0, _VREF_EXT_REF_P1N1, _VREF_AVDD_AVSS
    #                     _FIR_NONE, _FIR_5060Hz, _FIR_50Hz, _FIR_60Hz
    #
    def _set_vref_fir(self, vref, fir):
        if fir != 0x00 and self._rd_reg(1) >> 4: # sps_bits and higher mode bit (0 in normal or duty-cycle mode)
            raise ValueError('50 or 60 Hz FIR filter only at 5 or 20 SPS')
        self._wr_reg(2, vref | fir)

    def _calibrate(self):
        self._set_mux(3, 3)   # set mux to: ain_pos and ain_neg shortened to (AV_dd + AV_ss) / 2
        for _ in range(2):
            self._rd_data()
        self.offs = _N_Cal // 2
        for _ in range(_N_Cal):
            self.offs += self._rd_data()
        self.offs //= _N_Cal
        self._set_mux(self.ain_pos, self.ain_neg)
        for _ in range(2):  # being paranoid: the values may be corrupted by changing multiplexer setting ???
            self._rd_data()

# ----------------------- user routines start here

    def set_mux(self, ain_pos, ain_neg=4):
        self.ain_pos = ain_pos
        self.ain_neg = ain_neg
        self.cs.low()
        self._set_mux(ain_pos, ain_neg)
        self.cs.high()

    def calibrate(self):
        self.cs.low()
        self._calibrate()
        self.cs.high()

    def read(self):
        self.cs.low()
        r = self._rd_data() - self.offs
        self.cs.high()
        return r  #, self.waited_us

    def read_from(self, ain_pos, ain_neg=4):
        self.cs.low()
        self._set_mux(ain_pos, ain_neg)  # set multiplex control register
        return self.read()               # self.read() sets cs high afterwards

    def voltage(self):
        raw = self.read()
        return self.vref * raw / (self.gain * 0x800000)

    def voltage_from(self, a_pos, a_neg=4):
        raw = self.read_from(a_pos, a_neg)
        return self.vref * raw / (self.gain * 0x800000)

    def temperature(self):
        """return temperature in °Celsius"""
        self.cs.low()
        self._wr_reg(1, self._rd_reg(1) | 0x02)    #   sets temp mode bit
        temp = 0.03125 * (self._rd_data()>>10)
        self._wr_reg(1, self._rd_reg(1) & 0xfd)    # resets temp mode bit
        self.cs.high()
        return temp

    # for debugging, may be commented out later
    def config_dict(self):
        self.cs.low()
        reg = bytearray(4)
        r = self._rd_reg(0)
        reg[0] = r
        pga_byp = r & 0x01 > 0       # lower half of reg 0
        gain = 1 << ((r&0x0e)>>1)
        r >>= 4
        if 8 <= r < 12:              # upper half of reg 0
            ain_p = r - 8
            ain_n = 4
        else:
            for i in range(4):
                for j in range(4):
                    if r == _MUXTAB[i*4+j]:
                        break
            ain_p = i
            ain_n = j
        r = self._rd_reg(1)          # reg 1
        reg[1] = r
        temp_sens = r & 0x02 > 0
        cont_conv = r & 0x04 > 0
        cm = (r>>3) & 0x03
        cm = cm ^ (~cm & 0x02)>>1    # maps 0, 1, 2 to 1, 0, 2        
        sps =  _SRTAB[r>>5] * _SPSMULT[cm]
        r = self._rd_reg(2)          # reg 2
        reg[2] = r
        fir  = (r>>4) & 0x03
        vref = ['Internal', 'Ext_Ref_p0n0', 'Ext_Ref_p1n1', 'avdd_avss'][(r>>6) & 0x03]
        r = self._rd_reg(3)          # reg 3
        reg[3] = r
        drdy = r & 0x02 > 0
        self.cs.high()
        return {
            "PGA Bypass": pga_byp,
            "Gain": gain,
            "Ain_p": ain_p,
            "Ain_n": ain_n,
            "Temp Sensor Mode": temp_sens,
            "Cont Conversion": cont_conv,
            "Conversion Mode": ['Duty','Normal','Turbo'][cm],
            "Samples per Second": sps,
            "FIR Filter": ['None', '5060Hz', '50Hz', '60Hz'][fir],
            "Voltage Reference": vref,
            "Data Ready Mode": drdy,
            "Registers": ['0x{:02x}'.format(reg[0]), '0x{:02x}'.format(reg[1]), '0x{:02x}'.format(reg[2]), '0x{:02x}'.format(reg[3])]
        }

    # for debugging, may be commented out later
    def report_config(self):
        config_dict = self.config_dict()
        print("\n--- ads1220 config ---")
        for key, value in sorted(config_dict.items()):
            print("{0}: {1}".format(key, value))
        print("\n----------------------")

  